import io
import json
import os
import pathlib
import re
import sys
import typing
from contextlib import redirect_stdout
from datetime import date, timedelta
from itertools import zip_longest

from .utils import chunk_default

try:
    import click
    from click_aliases import ClickAliasedGroup
except ImportError:
    print("Missing dependencies for the CLI. Please `pip install aoc_helper[cli]`")
    exit(1)

from .data import (
    DATA_DIR,
    DEFAULT_YEAR,
    HEADERS,
    PRACTICE_DATA_DIR,
    RANK,
    URL,
    get_cookie,
)
from .interface import _estimate_practice_rank, _format_timedelta
from .interface import fetch as fetch_input
from .interface import submit as submit_answer
from .interface import validate_token as validate_token_

TEMPLATE = (pathlib.Path(__file__).parent / "day_template.py").read_text()

RANGE_REGEX = re.compile(r"(2[0-5]|1[0-9]|0?[1-9])-(2[0-5]|1[0-9]|0?[2-9])")


def parse_range(_, __, value: str) -> typing.List[int]:
    ranges = value.split(",")
    days: set[int] = set()
    for range_ in ranges:
        match = RANGE_REGEX.match(range_)
        if match:
            lb = int(match[1])
            ub = int(match[2]) + 1
            days |= set(range(lb, ub))
        elif range_.isnumeric() and 1 <= int(range_) <= 25:
            days.add(int(range_))
        elif range_ == "all":
            days = set(range(1, 26))
        else:
            raise click.BadParameter(
                "every part must be a single day, a range of days in the form "
                "a-b, or the word 'all'"
            )
    return sorted(days)


@click.group(cls=ClickAliasedGroup)
def cli():
    pass


@cli.command(aliases=["f", "get", "get-input", "download"])
@click.argument("day", type=int)
@click.option("--year", type=int, default=DEFAULT_YEAR)
def fetch(day: int, year: int):
    """Fetch and print the input for day DAY of --year"""
    print(fetch_input(day, year))


@cli.command(aliases=["read-puzzle", "puzzle"])
@click.argument("day", type=int)
@click.option("--year", type=int, default=DEFAULT_YEAR)
@click.option(
    "-c",
    "--colour",
    "--color",
    type=click.Choice(["auto", "always", "never"]),
    default="auto",
)
def read(day: int, year: int, colour: typing.Literal["auto", "always", "never"]):
    """Read the puzzle for day DAY or --year in your terminal"""
    import sys
    from os import getenv

    import requests
    from bs4 import BeautifulSoup, Tag

    try:
        from rich.console import Console, ConsoleOptions, RenderResult
        from rich.markdown import TextElement
        from rich.segment import Segment
    except ImportError:
        print(
            "Missing dependency rich. Please `pip install rich`,"
            " `pip install aoc_helper[full]` or `pip install aoc_helper[fancy]`",
            file=sys.stderr,
        )
        return
    puzzle_info = requests.get(
        URL.format(year=year, day=day), headers=HEADERS, cookies=get_cookie()
    )
    soup = BeautifulSoup(puzzle_info.text, "html.parser")
    puzzle: Tag = soup.find("main")  # type: ignore
    for emphasis in puzzle.find_all("em"):
        emphasis.string.replace_with(f"[bold gold1]{emphasis.text}[/]")
    terminal = Console()
    if colour == "auto":
        pager = getenv("PAGER") or ""
        colour = (
            "always"
            if terminal.is_terminal
            and any(
                flag in pager
                for flag in ("-r", "-R", "--raw-control-chars", "--RAW-CONTROL-CHARS")
            )
            else "never"
        )

    class CodeBlock(TextElement):
        def __init__(self, text: str):
            self.text = text

        def __rich_console__(
            self, console: Console, options: ConsoleOptions
        ) -> RenderResult:
            lines = self.text.split("\n")
            line_count = len(lines)
            width = len(str(line_count))
            render_options = options.update(width=console.width - 7 - width)
            for line_no, line in enumerate(lines, 1):
                inner_lines = console.render_lines(line, render_options)
                for i, line in enumerate(inner_lines):
                    yield Segment("    ")
                    yield Segment(
                        f"{line_no:>{width}} │ " if i == 0 else " " * (width) + " │ "
                    )
                    yield from line
                    yield Segment("\n")

    class BulletItem(TextElement):
        def __init__(self, text: str):
            self.text = text

        def __rich_console__(
            self, console: Console, options: ConsoleOptions
        ) -> RenderResult:
            render_options = options.update(width=console.width - 2)
            lines = console.render_lines(self.text, render_options)
            for i, line in enumerate(lines):
                yield Segment("- " if i == 0 else "  ")
                yield from line
                yield Segment("\n")

    class NumberedItem(TextElement):
        def __init__(self, text: str, number: int, width: int):
            self.text = text
            self.number = number
            self.width = width

        def __rich_console__(
            self, console: Console, options: ConsoleOptions
        ) -> RenderResult:
            render_options = options.update(width=console.width - self.width - 2)
            lines = console.render_lines(self.text, render_options)
            for i, line in enumerate(lines):
                yield (
                    f"{self.number:>{self.width}}. "
                    if i == 0
                    else (" " * self.width + "  ")
                )
                yield from line
                yield "\n"

    with terminal.pager(styles=colour == "always") as pager:
        first = True
        for el in puzzle.children:
            if not isinstance(el, Tag):
                continue
            if el.name == "article":
                for part_el in el.children:
                    if not isinstance(part_el, Tag):
                        continue
                    if not first:
                        terminal.print()
                    first = False
                    if part_el.name == "h2":
                        terminal.rule(
                            "[bold gold1 underline]" + part_el.text.strip("- "),
                            style="bold gold1",
                        )
                    elif part_el.name == "p":
                        terminal.print(part_el.text)
                    elif part_el.name == "pre":
                        terminal.print(CodeBlock(part_el.text.strip("\n")))
                    elif part_el.name == "ul":
                        for li in part_el.find_all("li"):
                            terminal.print(BulletItem(li.text))
                    elif part_el.name == "ol":
                        lis = list(part_el.find_all("li"))
                        width = len(str(len(lis)))
                        for i, li in enumerate(lis, 1):
                            terminal.print(NumberedItem(li.text, i, width))

            elif el.name == "p":
                if el.text.startswith("Your puzzle answer was"):
                    terminal.print()
                    terminal.print()
                    terminal.print(el.text)
                    terminal.print()
                elif el.text.startswith("Both parts of this puzzle are complete!"):
                    terminal.print(f"[bold gold1]{el.text}[/]")
                    return


@cli.command(aliases=["s", "send"])
@click.argument("day", type=int)
@click.argument("part", type=click.Choice(["1", "2"]))
@click.argument("answer")
@click.option("--year", type=int, default=DEFAULT_YEAR)
@click.option("--practice", is_flag=True)
def submit(
    day: int, part: typing.Literal["1", "2"], answer: str, year: int, practice: bool
):
    """Submit the answer for day DAY part PART of --year"""
    _ = practice  # used via sys.argv
    submit_answer(day, int(part), answer, year)


@cli.command(aliases=["t", "create"])
@click.argument("days", callback=parse_range)
@click.option("--year", type=int, default=DEFAULT_YEAR)
def template(days: typing.List[int], year: int):
    """Generate an answer stub for every day of DAYS in --year"""
    for day in days:
        print(f"Generating day_{day:0>2}.py")
        pathlib.Path(f"day_{day:0>2}.py").write_text(
            TEMPLATE.format(day=day, year=year)
        )


@cli.command(aliases=["get-browser", "set-browser"])
@click.argument("state", type=bool, default=None, required=False)
def browser(state: typing.Optional[bool]):
    """Enable, disable, or check browser automation"""
    file = DATA_DIR / ".nobrowser"
    if state is None:
        print(f"Web browser automation is {'dis' if file.exists() else 'en'}abled.")
    elif state:
        file.unlink(True)
        print("Enabled web browser automation")
    else:
        file.touch()
        print("Disabled web browser automation")


@cli.command(
    aliases=[
        "clear",
        "purge",
        "delete",
        "clear-cache",
        "clean-cache",
        "purge-cache",
        "delete-cache",
    ]
)
@click.argument("days", callback=parse_range)
@click.argument("year", type=int, default=DEFAULT_YEAR)
@click.option(
    "--type",
    type=click.Choice(
        ["input", "submissions", "solutions", "1", "2", "tests", "practice", "all"]
    ),
    help="What to delete",
    default="input",
)
def clean(days: typing.List[int], year: int, type: str):
    """Clean the cached --type data for DAYS of YEAR"""
    for day in days:
        if type in ("input", "all"):
            (DATA_DIR / f"{year}" / f"{day}.in").unlink(True)
        if type in ("submissions", "all"):
            file = DATA_DIR / f"{year}" / f"{day}" / "submissions.json"
            if (
                not file.exists()
                or not RANK.search(file.read_text())
                or click.confirm(
                    f"Are you sure you want to delete your submissions for {year} day"
                    f" {day}? Your cached rank will be forgotten"
                )
            ):
                file.unlink(True)
        if type in ("practice", "all"):
            folder = PRACTICE_DATA_DIR / f"{year}" / f"{day}"
            if not folder.exists() or click.confirm(
                f"Are you sure you want to delete your practice data for {year} day"
                f" {day}? {len(os.listdir(folder))} entries will be forgotten"
            ):
                for file in os.listdir(folder):
                    (folder / file).unlink(True)
                folder.rmdir()
        if type in ("solutions", "all", "1"):
            (DATA_DIR / f"{year}" / f"{day}" / "1.solution").unlink(True)
        if type in ("solutions", "all", "2"):
            (DATA_DIR / f"{year}" / f"{day}" / "2.solution").unlink(True)
        if type in ("tests", "all"):
            (DATA_DIR / f"{year}" / f"{day}" / "tests.json").unlink(True)


@cli.command(
    name="visualise-cache",
    aliases=["cache", "visualize-cache", "list-cache", "view-cache"],
)
@click.option(
    "-c",
    "--colour",
    "--color",
    type=click.Choice(["auto", "always", "never"]),
    default="auto",
)
@click.option(
    "--validate-token/--no-validate-token",
    default=True,
)
@click.option(
    "--update-token-on-invalid/--no-update-token-on-invalid",
    default=False,
    help=(
        "Whether to prompt for a new token if the current one is invalid."
        " Only takes effect if --validate-token is set"
    ),
)
def visualise_cache(
    colour: typing.Literal["auto", "always", "never"],
    validate_token: bool,
    update_token_on_invalid: bool,
):
    """Get a visual overview of the aoc_helper cache"""
    from .formatting import print

    try:
        from rich.console import Console
    except ImportError:
        if colour != "never":
            print(
                "Missing dependency rich. Colour will be disabled. Please",
                "`pip install rich`, `pip install aoc_helper[full]`, or",
                "`pip install aoc_helper[fancy]`",
                file=sys.stderr,
            )

        def rule(title: str):
            width = 73 - 4 - len(title)
            left = width // 2
            right = width - left
            print(f" {'-' * left} {title} {'-' * right}")

        real_width = 73
    else:
        terminal = Console()
        real_width = terminal.width if terminal.is_terminal else 73
        terminal.width = 73
        if colour != "never":
            if colour == "auto":
                colour = "always" if terminal.is_terminal else "never"
            if colour == "always":
                Console._environ["FORCE_COLOR"] = "1"  # type: ignore
            rule = terminal.rule
        else:

            def rule(title: str):
                width = 73 - 2 - len(title)
                left = width // 2
                right = width - left
                print(f"{'=' * left} {title} {'=' * right}")

    if colour == "always":
        from .formatting import GREEN, RED, RESET, YELLOW
    else:
        GREEN = RED = YELLOW = RESET = ""

    token = get_cookie(missing_ok=not update_token_on_invalid)
    cached_years = sorted(DATA_DIR.iterdir())
    did_print = False

    if token is not None:
        valid = None
        if validate_token:
            valid = validate_token_(update_token_on_invalid)
        if colour == "always" or valid is not False:
            print(
                {
                    None: f"{YELLOW}Token (not validated){RESET}",
                    True: f"{GREEN}Token{RESET}",
                    False: f"{RED}Token{RESET}",
                }[valid]
            )
            did_print = True
    elif colour == "always":
        print(f"{RED}Token{RESET}")
        did_print = True

    def format(
        val: str, exists: bool, success_colour: str = GREEN, fail_colour: str = RED
    ):
        if colour == "always":
            text_col = success_colour if exists else fail_colour
            return f"{text_col}{val}{text_col and RESET}"
        else:
            return val if exists else (" " * len(val))

    years = [year for year in cached_years if year.is_dir() and year.name.isnumeric()]
    # input   tests
    # solutions 1 2
    # practice data
    days = range(1, 26)
    blocks_that_fit = max((real_width + 4) // 77, 1)

    blocks = []
    for year in years:
        with io.StringIO() as buf, redirect_stdout(buf):
            rule(year.name)
            has_input = [(year / f"{day}.in").exists() for day in days]
            has_tests = [(year / f"{day}" / "tests.json").exists() for day in days]
            has_solution_1 = [(year / f"{day}" / "1.solution").exists() for day in days]
            has_solution_2 = [(year / f"{day}" / "2.solution").exists() for day in days]
            has_any_solution = [a or b for a, b in zip(has_solution_1, has_solution_2)]
            has_practice = [
                (PRACTICE_DATA_DIR / year.name / f"{day}").exists() for day in days
            ]
            for block in range(5):
                print(
                    *(f"{day:^13}" for day in range(block * 5 + 1, block * 5 + 6)),
                    sep="  ",
                )
                print(
                    *(
                        f"{format('Input', has_input[day])}   {format('Tests', has_tests[day])}"
                        for day in range(block * 5, block * 5 + 5)
                    ),
                    sep="  ",
                )
                print(
                    *(
                        f"{format('Solutions', has_any_solution[day], '', '')}"
                        f" {format('1', has_solution_1[day])}"
                        f" {format('2', has_solution_2[day])}"
                        for day in range(block * 5, block * 5 + 5)
                    ),
                    sep="  ",
                )
                print(
                    *(
                        format("Practice data", has_practice[day])
                        for day in range(block * 5, block * 5 + 5)
                    ),
                    sep="  ",
                )
            blocks.append(buf.getvalue())

    import builtins

    for blocks in chunk_default(blocks, blocks_that_fit, ""):
        if did_print:
            print()
        blocks = [block.splitlines() for block in blocks]
        for lines in zip_longest(*blocks, fillvalue=""):
            builtins.print("    ".join(lines))


@cli.command(name="validate-token", aliases=["token", "check-token"])
def validate_token():
    """
    Validate the stored session token. Will prompt for a new token if the
    current one is invalid.
    """
    if validate_token_():
        print("Token is valid")


@cli.command(name="practice-results")
@click.argument("day", type=int)
@click.option("--year", type=int, default=DEFAULT_YEAR)
def practice_results(day: int, year: int):
    """Show all practice results for day DAY of --year"""
    import locale

    locale.setlocale(
        locale.LC_TIME, ""
    )  # https://github.com/python/cpython/issues/73643
    folder = PRACTICE_DATA_DIR / f"{year}" / f"{day}"
    if not folder.exists():
        print("No practice results found")
        return

    def format_result(result: typing.Optional[typing.Tuple[int, int, int]]):
        if not result:
            return "no rank"
        estimated, best, worst = result
        if best == worst:
            return f"rank {best}"
        if worst > 100:
            worst = "100+"
        return f"approximately rank {estimated} - {best} to {worst}"

    for file in sorted(folder.iterdir()):
        attempt_year, attempt_month, attempt_day = map(int, file.stem.split("-"))
        attempt_date = date(attempt_year, attempt_month, attempt_day).strftime("%x")
        results: typing.List[float] = json.loads(file.read_text())
        if len(results) == 1:
            solve_time = timedelta(seconds=results[0])
            result = format_result(_estimate_practice_rank(day, 1, year, solve_time))
            print(
                f"{attempt_date} - Part 1: {_format_timedelta(solve_time)} ({result}),"
                " Part 2: (unsolved)"
            )
        elif len(results) == 2:
            solve_time_1 = timedelta(seconds=results[0])
            solve_time_2 = timedelta(seconds=results[1])
            result_1 = format_result(
                _estimate_practice_rank(day, 1, year, solve_time_1)
            )
            result_2 = format_result(
                _estimate_practice_rank(day, 2, year, solve_time_2)
            )
            print(
                f"{attempt_date} - "
                f"Part 1: {_format_timedelta(solve_time_1)} ({result_1}), "
                f"Part 2: {_format_timedelta(solve_time_2)} ({result_2})"
            )



import datetime
import json
import pathlib
import sys
import typing
import webbrowser
from warnings import warn

import requests
from bs4 import BeautifulSoup as Soup

from .data import (
    DATA_DIR,
    DEFAULT_YEAR,
    HEADERS,
    LEADERBOARD_URL,
    PRACTICE_DATA_DIR,
    RANK,
    TODAY,
    URL,
    WAIT_TIME,
    get_cookie,
)
from .formatting import BLUE, GOLD, GREEN, RED, RESET, YELLOW, print, wait, work

T = typing.TypeVar("T")
U = typing.TypeVar("U")


def _open_page(page: str) -> None:
    """Open the page if the user hasn't opted out"""
    if not (DATA_DIR / ".nobrowser").exists():
        webbrowser.open(page)


def _make(folder: pathlib.Path) -> None:
    """Create folder if it doesn't exist."""
    if not folder.exists():
        folder.mkdir(parents=True)


def _pretty_print(message: str) -> None:
    """Analyse and print message"""
    if message.startswith("That's the"):
        print(GREEN + message + RESET)
    elif message.startswith("You don't"):
        print(YELLOW + message + RESET)
    elif message.startswith("That's not"):
        print(RED + message + RESET)
    elif message.startswith("You got rank"):
        print(GOLD + message + RESET)
    else:
        raise ValueError("Failed to parse response.")


def validate_token(allow_prompt: bool = True) -> bool:
    cookie = get_cookie(not allow_prompt)
    if cookie is None:
        return False
    resp = requests.get(
        URL.format(day=1, year=2015) + "/input",
        cookies=get_cookie(),
        headers=HEADERS,
    )
    if not resp.ok:
        if resp.status_code != 400:
            raise ValueError("Received bad response")
        if not allow_prompt:
            return False
        token_file = DATA_DIR / "token.txt"
        print(f"{RED}Your token has expired. Please enter your new" f" token.{RESET}")
        token = input(">>> ")
        token_file.write_text(token)
        return validate_token(True)
    return True


def fetch(day: int = TODAY, year: int = DEFAULT_YEAR, never_print: bool = False) -> str:
    """Fetch and return the input for `day` of `year`.

    If `--practice` is provided on the command line, pretend that today is the
    day of the puzzle and wait for puzzle unlock accordingly. 'today' is
    determined by UTC; from 0:00 to 5:00 UTC, this puzzle will block until 5:00
    UTC - after that, until 0:00 UTC the next day, input fetching will be
    instant.

    All inputs are cached in `aoc_helper.DATA_DIR`."""
    import sys

    day_ = str(day)
    year_ = str(year)

    _make(DATA_DIR / year_)
    input_path = DATA_DIR / year_ / (day_ + ".in")

    if input_path.exists():
        should_print = False
        if "--practice" in sys.argv:
            now = datetime.datetime.utcnow()
            unlock = datetime.datetime(now.year, now.month, now.day, 5)
            if now < unlock:
                should_print = True
                wait(
                    f"{YELLOW}Waiting for puzzle unlock...{RESET}",
                    (unlock - now).total_seconds(),
                )
                print(GREEN + "Fetching input!" + RESET)
                _open_page(URL.format(day=day, year=year))
        input_data = input_path.read_text()
        if "--practice" in sys.argv and should_print:
            print(input_data)
        return input_data
    else:
        unlock = datetime.datetime(year, 12, day, 5)
        now = datetime.datetime.utcnow()
        if "--practice" in sys.argv:
            unlock = unlock.replace(year=now.year, month=now.month, day=now.day)
        if now < unlock:
            if day == 1:
                # If we're waiting for the first puzzle, the user's token may
                # have expired. We should check that now.
                validate_token()
                now = datetime.datetime.utcnow()
            wait(
                f"{YELLOW}Waiting for puzzle unlock...{RESET}",
                (unlock - now).total_seconds(),
            )
            print(GREEN + "Fetching input!" + RESET)
            _open_page(URL.format(day=day, year=year))
        resp = requests.get(
            URL.format(day=day, year=year) + "/input",
            cookies=get_cookie(),
            headers=HEADERS,
        )
        if not resp.ok:
            if resp.status_code == 400:
                token_file = DATA_DIR / "token.txt"
                print(
                    f"{RED}Your token has expired. Please enter your new token.{RESET}"
                )
                token = input(">>> ")
                token_file.write_text(token)
                return fetch(day, year, never_print)
            raise ValueError("Received bad response")
        data = resp.text.strip("\n")
        input_path.write_text(data)
        if not never_print:
            print(data)
        return data


def _load_leaderboard_times(
    day: int, year: int = DEFAULT_YEAR
) -> typing.Tuple[typing.List[datetime.timedelta], typing.List[datetime.timedelta]]:
    day_ = str(day)
    year_ = str(year)

    day_dir = DATA_DIR / year_ / day_
    _make(day_dir)

    # Load cached leaderboards
    leaderboards = day_dir / "leaderboards.json"
    if leaderboards.exists():
        with leaderboards.open() as f:
            # ([seconds], [seconds])
            data: typing.List[typing.List[int]] = json.load(f)
            return [datetime.timedelta(seconds=t) for t in data[0]], [datetime.timedelta(seconds=t) for t in data[1]]  # type: ignore
    else:
        leaderboard_page = requests.get(
            LEADERBOARD_URL.format(day=day, year=year), headers=HEADERS
        )
        soup = Soup(leaderboard_page.text, "html.parser")
        times = soup.select(".leaderboard-entry")
        part_1_times: typing.List[datetime.timedelta] = []
        part_2_times: typing.List[datetime.timedelta] = []
        in_part_2 = False
        for leaderboard_time in times:
            if leaderboard_time.span.text == "  1)":  # type: ignore
                in_part_2 = not in_part_2
            time_to_solve = datetime.datetime.strptime(
                leaderboard_time.select_one(".leaderboard-time").text,  # type: ignore
                "%b %d  %H:%M:%S",
            ) - datetime.datetime(1900, 12, day)
            if in_part_2:
                part_2_times.append(time_to_solve)
            else:
                part_1_times.append(time_to_solve)
        if not part_1_times:
            # no part 2 leaderboard, so boards were read in backwards
            part_2_times, part_1_times = part_1_times, part_2_times
        if len(part_1_times) == len(part_2_times) == 100:
            # both leaderboards are full, cache them
            with leaderboards.open("w") as f:
                json.dump(
                    (
                        [t.total_seconds() for t in part_1_times],
                        [t.total_seconds() for t in part_2_times],
                    ),
                    f,
                )
        return part_1_times, part_2_times


def _practice_result_for(day: int, year: int) -> typing.List[int]:
    practice_data_dir = PRACTICE_DATA_DIR / str(year) / str(day)
    _make(practice_data_dir)
    try:
        with open(
            practice_data_dir
            / f"{datetime.datetime.utcnow().year:04}-{datetime.datetime.utcnow().month:02}-{datetime.datetime.utcnow().day:02}.json",
            "r",
        ) as f:
            return json.load(f)
    except FileNotFoundError:
        return []


def _calculate_practice_result(day: int, part: int, year: int) -> None:
    if "--practice" not in sys.argv:
        return
    now = datetime.datetime.utcnow()
    solve_time = datetime.timedelta(
        hours=now.hour - 5,
        minutes=now.minute,
        seconds=now.second,
        microseconds=now.microsecond,
    )
    practice_data_dir = PRACTICE_DATA_DIR / str(year) / str(day)
    _make(practice_data_dir)

    filename = f"{now.year:04}-{now.month:02}-{now.day:02}.json"

    try:
        with open(practice_data_dir / filename) as f:
            data: typing.List[float] = json.load(f)
    except (json.decoder.JSONDecodeError, FileNotFoundError):
        data = []
    with open(practice_data_dir / filename, "w") as f:
        data.append(solve_time.total_seconds())
        json.dump(data, f)
    _report_practice_result(day, part, year, solve_time)


def _estimate_practice_rank(
    day: int, part: int, year: int, solve_time: datetime.timedelta
) -> typing.Optional[typing.Tuple[int, int, int]]:
    import bisect

    leaderboard = _load_leaderboard_times(day, year)[part - 1]
    # aoc truncates solve times, so we do too for the purpose of sorting
    truncated_solve_time = datetime.timedelta(seconds=int(solve_time.total_seconds()))
    best_possible_rank = bisect.bisect_left(leaderboard, truncated_solve_time) + 1
    worst_possible_rank = bisect.bisect_right(leaderboard, truncated_solve_time) + 1
    if best_possible_rank > 100:
        return None
    span = worst_possible_rank - best_possible_rank
    approx_rank = best_possible_rank + round(span * solve_time.microseconds / 1_000_000)
    return approx_rank, best_possible_rank, worst_possible_rank


def _format_timedelta(solve_time: datetime.timedelta) -> str:
    minutes, seconds = divmod(solve_time.seconds, 60)
    hours, minutes = divmod(minutes, 60)
    if hours > 0:
        return f"{hours:02}:{minutes:02}:{seconds:02}"
    else:
        return f"{minutes:02}:{seconds:02}.{solve_time.microseconds // 10_000:02}"


def _report_practice_result(
    day: int, part: int, year: int, solve_time: datetime.timedelta
) -> None:
    print(
        f"{GREEN}You solved the puzzle in"
        f" {BLUE}{_format_timedelta(solve_time)}{GREEN}!{RESET}"
    )
    result = _estimate_practice_rank(day, part, year, solve_time)
    if not result:
        print(f"{YELLOW}You would not have achieved a leaderboard position.{RESET}")
    else:
        likely_rank, best_possible_rank, worst_possible_rank = result
        if best_possible_rank == worst_possible_rank:
            print(f"{GOLD}You would have achieved rank {best_possible_rank}!{RESET}")
        else:
            if worst_possible_rank > 100:
                worst_possible_rank = "100+"
            print(
                f"{GOLD}You would have achieved approximately rank"
                f" {likely_rank} ({best_possible_rank} to"
                f" {worst_possible_rank})!{RESET}"
            )


def submit(day: int, part: int, answer: typing.Any, year: int = DEFAULT_YEAR) -> None:
    """Submit a solution.

    Submissions are cached; submitting an already submitted solution will return the
    previous response.
    """
    day_ = str(day)
    year_ = str(year)
    part_ = str(part)
    answer_ = str(answer)

    submission_dir = DATA_DIR / year_ / day_
    _make(submission_dir)

    # Load cached solutions
    submissions = submission_dir / "submissions.json"
    if submissions.exists():
        with submissions.open() as f:
            solutions = json.load(f)
    else:
        solutions = {"1": {}, "2": {}}

    # Check if solved
    solution_file = submission_dir / f"{part}.solution"
    if solution_file.exists():
        solution = solution_file.read_text()
        if "--practice" in sys.argv:
            if solution == answer_:
                _calculate_practice_result(day, part, year)
            else:
                print(
                    f"{RED}Submitted {BLUE}{answer_}{RESET}; that's not the right"
                    f" answer.{RESET}\n"
                )
            return
        print(
            f"Day {BLUE}{day}{RESET} part {BLUE}{part}{RESET} "
            "has already been solved.\nThe solution was: "
            f"{BLUE}{solution}{RESET}"
        )
        _print_rank(solutions[part_][solution])
        return

    # Check if answer has already been submitted
    if answer_ in solutions[part_]:
        print(
            f"{YELLOW}Solution: {BLUE}{answer}{RESET} to part "
            f"{BLUE}{part}{RESET} has already been submitted.\n"
            f"Response was:{RESET}"
        )
        return _pretty_print(solutions[part_][answer_])

    while True:
        print(
            f"Submitting {BLUE}{answer}{RESET} as the solution to part "
            f"{BLUE}{part}{RESET}..."
        )
        resp = requests.post(
            url=URL.format(day=day, year=year) + "/answer",
            cookies=get_cookie(),
            data={"level": part_, "answer": answer_},
            headers=HEADERS,
        )
        if not resp.ok:
            if resp.status_code == 400:
                token_file = DATA_DIR / "token.txt"
                token = input(
                    "Your token has expired. Please enter your new token\n>>> "
                )
                token_file.write_text(token)
                return submit(day, part, answer, year)
            raise ValueError("Received bad response")

        article = Soup(resp.text, "html.parser").article
        assert article is not None
        msg = article.text

        if msg.startswith("You gave"):
            print(RED + msg + RESET)
            wait_match = WAIT_TIME.search(msg)
            assert wait_match is not None
            pause = 60 * int(wait_match[1] or 0) + int(wait_match[2])
            wait(
                f"{YELLOW}Waiting {BLUE}{pause}{RESET} seconds to retry...{RESET}",
                pause,
            )
        else:
            break

    if msg.startswith("That's the"):
        _print_rank(msg)
        solution_file.write_text(answer_)
        _calculate_practice_result(day, part, year)
        if part == 1:
            if not resp.url.endswith("#part2"):
                resp.url += "#part2"  # scroll to part 2
            _open_page(resp.url)  # open part 2 in the user's browser
    else:
        _pretty_print(msg)

    # Cache submission
    solutions[part_][answer_] = msg
    with submissions.open("w") as f:
        json.dump(solutions, f)


def _print_rank(msg: str) -> None:
    match = RANK.search(msg)
    if match:
        _pretty_print(f"You got rank {match.group(1)} for this puzzle")
    else:
        _pretty_print(msg)


def submit_25(year: str):
    print(f"{GREEN}Finishing Advent of Code {BLUE}{year}{RESET}!{RESET}")
    resp = requests.post(
        url=URL.format(day="25", year=year) + "/answer",
        cookies=get_cookie(),
        data={"level": "2", "answer": "0"},
        headers=HEADERS,
    )
    if not resp.ok:
        if resp.status_code == 400:
            token_file = DATA_DIR / "token.txt"
            token = input("Your token has expired. Please enter your new token\n>>> ")
            token_file.write_text(token)
            return submit_25(year)
        raise ValueError("Received bad response")

    print("Response from the server:")
    article = Soup(resp.text, "html.parser").article
    assert article is not None
    print(article.text.strip())
    if len(_practice_result_for(25, int(year))) < 2:
        _calculate_practice_result(25, 2, int(year))


def lazy_submit(
    day: int,
    solution: typing.Callable[[U], typing.Any],
    data: U,
    year: int = DEFAULT_YEAR,
) -> None:
    """Run the function only if we haven't seen a solution.

    Will also run the solution if `--force-run` or `--practice` is passed on the
    command line.

    solution is expected to be named 'part_one' or 'part_two'
    """
    import sys

    part = 1 if solution.__name__ == "part_one" else 2
    submission_dir = DATA_DIR / str(year) / str(day)
    if day == 25 and part == 2:
        # Don't try to submit part 2 if part 1 isn't solved
        if (submission_dir / "1.solution").exists():
            submit_25(str(year))
        else:
            return
    solution_file = submission_dir / f"{part}.solution"
    # Check if solved
    if (
        solution_file.exists()
        and "--force-run" not in sys.argv
        and (
            "--practice" not in sys.argv  # not in practice mode
            or len(_practice_result_for(day, year))
            >= part  # or solved today in practice
        )
    ):
        # Load cached solutions
        submissions = submission_dir / "submissions.json"
        with submissions.open() as f:
            solutions = json.load(f)

        solution_ = solution_file.read_text()
        print(
            f"Day {BLUE}{day}{RESET} part {BLUE}{part}{RESET} "
            "has already been solved.\nThe solution was: "
            f"{BLUE}{solution_}{RESET}"
        )
        _print_rank(solutions[str(part)][solution_])
    else:
        answer = work(
            f"{YELLOW}Running part"
            f" {RESET}{BLUE}{part}{RESET}{YELLOW} solution...{RESET}",
            solution,
            data,
        )
        if answer is not None:
            submit(day, part, answer, year)


def get_sample_input(
    day: int, part: int, year: int = DEFAULT_YEAR
) -> typing.Optional[typing.Tuple[str, str]]:
    """Retrieves the example input and answer for the corresponding AOC challenge."""
    testing_dir = DATA_DIR / str(year) / str(day)
    _make(testing_dir)
    testing_file = testing_dir / "tests.json"

    if testing_file.exists():
        test_info: typing.Dict[str, typing.Optional[typing.Tuple[str, str]]] = (
            json.loads(testing_file.read_text())
        )
    else:
        test_info: typing.Dict[str, typing.Optional[typing.Tuple[str, str]]] = {}

    if str(part) in test_info:
        return test_info[str(part)]

    resp = requests.post(
        url=URL.format(day=day, year=year), cookies=get_cookie(), headers=HEADERS
    )
    soup = Soup(resp.text, "html.parser")

    example_test_inputs = []
    # Find the example test input for that day.
    for possible_test_input in soup.find_all("pre"):
        preceding_text = (
            possible_test_input.previous_element.previous_element.text.lower()
        )
        if (
            "for example" in preceding_text
            or "consider" in preceding_text
            or "given" in preceding_text
        ) and ":" in preceding_text:
            example_test_inputs.append(possible_test_input.text.strip())

    if not example_test_inputs:
        test_info[str(part)] = None
        testing_file.write_text(json.dumps(test_info))
        warn(
            f"An issue occurred while fetching test data for {year} day"
            f" {day} part {part}. You may either ignore this message, or pass"
            " custom test data to lazy_test.",
            RuntimeWarning,
        )
        return

    try:
        test_input = example_test_inputs[-1]

        # Attempt to retrieve answer to said example data from puzzle part.
        current_part = soup.find_all("article")[part - 1]
        last_sentence = current_part.find_all("p")[-2]
        answer = last_sentence.find_all("code")[-1]
    except IndexError:
        test_info[str(part)] = None
        testing_file.write_text(json.dumps(test_info))
        warn(
            f"An issue occurred while fetching test data for {year} day"
            f" {day} part {part}. You may either ignore this message, or pass"
            " custom test data to lazy_test.",
            RuntimeWarning,
        )
        return
    if not answer.em:
        try:
            answer = last_sentence.find_all("em")[-1]
        except IndexError:
            pass

    try:
        answer = answer.text.strip("\n").split()[-1]
    except IndexError:
        test_info[str(part)] = None
        testing_file.write_text(json.dumps(test_info))
        warn(
            f"An issue occurred while fetching test data for {year} day"
            f" {day} part {part}. You may either ignore this message, or pass"
            " custom test data to lazy_test.",
            RuntimeWarning,
        )
        return

    test_data = test_input, answer
    test_info[str(part)] = test_data
    testing_file.write_text(json.dumps(test_info))
    return test_data


def _test(part: int, answer: str, expected_answer: str) -> None:
    assert answer == expected_answer, (
        f"The expected answer for the example test input was {expected_answer} but"
        f" your answer was {answer}."
    )
    print(
        f"{GREEN}Test for part {BLUE}{part}{RESET} succeeded!"
        f" The answer for part {BLUE}{part}{RESET} with the test data was:"
        f" {BLUE}{answer}{RESET}{RESET}"
    )


def lazy_test(
    day: int,
    parse: typing.Callable[[str], T],
    solution: typing.Callable[[T], typing.Any],
    year: int = DEFAULT_YEAR,
    test_data: typing.Optional[typing.Tuple[str, typing.Any]] = None,
) -> None:
    """Test the function with AOC's example data only if we haven't tested it already.

    Solution is expected to be named 'part_one' or 'part_two'
    """
    part = 1 if solution.__name__ == "part_one" else 2
    testing_dir = DATA_DIR / str(year) / str(day)
    _make(testing_dir)

    # If this is part 2, skip fetching/running tests if part 1 hasn't been submitted
    if part == 2 and not (testing_dir / "1.solution").exists():
        return

    # If this part has been submitted, skip running tests
    if not (testing_dir / f"{part}.solution").exists():
        if test_data is None:  # No test data passed (most common)
            test_data = get_sample_input(day, part, year)
            if test_data is None:  # No test data scraped (uncommon)
                return
        test_input, test_answer = test_data

        answer = work(
            f"{YELLOW}Running the test for part {BLUE}{part}{RESET} solution...{RESET}",
            solution,
            parse(test_input),
        )
        if answer is not None:
            _test(part, str(answer).strip(), str(test_answer).strip())



import builtins
import time
import typing

T = typing.TypeVar("T")
U = typing.TypeVar("U")

try:
    from rich import print, progress
except ImportError:
    from builtins import print  # suppress a Pylance warning

    RED = ""
    YELLOW = ""
    GREEN = ""
    BLUE = ""
    GOLD = ""
    RESET = ""

    def wait(msg: str, secs: float) -> None:
        print(msg)
        time.sleep(secs)

    def work(msg: str, worker: typing.Callable[[U], T], data: U) -> T:
        print(msg)
        return worker(data)

else:
    RED = "[red]"
    YELLOW = "[yellow]"
    GREEN = "[green]"
    BLUE = "[blue]"
    GOLD = "[gold1]"
    RESET = "[/]"

    def wait(msg: str, secs: float) -> None:
        for _ in progress.track(
            builtins.range(int(10 * secs)),
            description=msg,
            show_speed=False,
            transient=True,
        ):
            time.sleep(0.1)

    def _rich_work(msg: str, worker: typing.Callable[[U], T], data: U) -> T:
        with progress.Progress(
            progress.TextColumn("{task.description}"),
            progress.SpinnerColumn(),
            progress.TimeElapsedColumn(),
            transient=True,
        ) as bar:
            task = bar.add_task(msg)
            val = worker(data)
            bar.advance(task)
            return val

    work = _rich_work

__all__ = ("RED", "YELLOW", "GREEN", "BLUE", "GOLD", "RESET", "wait", "work", "print")
